#include "TCP.h"
#include "utility/w5500.h"
#include "utility/socket.h"

ConnectionStatusSimple EthernetMaketClient::connectNonBlock(IPAddress ip, uint16_t port) {
 
  if (_sock != MAX_SOCK_NUM) {
    // Serial.println(F("connectNonBlock"));
    uint8_t status_ = status();
    switch (status_) {
      case SnSR::CLOSED:
        // Serial.println(F("CLOSED"));
        _sock = MAX_SOCK_NUM;
        return CONNECT_FAIL;
        break;
      case SnSR::ESTABLISHED:
        // Serial.println(F("ESTABLISHED"));
        return CONNECT_SUCCESS;
        break;
      default:
        // Serial.print(F("status "));
        // Serial.println(status_,HEX);
        return CONNECT_CONNECTING;
        break;
    }
  }
  for (int i = 0; i < MAX_SOCK_NUM; i++) {
    uint8_t s = w5500.readSnSR(i);
    if (s == SnSR::CLOSED || s == SnSR::FIN_WAIT || s == SnSR::CLOSE_WAIT) {
      _sock = i;
      break;
    }
  }

  if (_sock == MAX_SOCK_NUM)
    return CONNECT_FAIL;

  _srcport++;
  if (_srcport == 0) _srcport = 1024;
  socket(_sock, SnMR::TCP, _srcport, 0);

  if (!::connect(_sock, rawIPAddress(ip), port)) {
    _sock = MAX_SOCK_NUM;
    return CONNECT_FAIL;
  }
  
  return CONNECT_CONNECTING;
}

void EthernetMaketClient::disconnect(){
  if (_sock == MAX_SOCK_NUM)
    return;
    ::disconnect(_sock);
}

void EthernetMaketClient::close(){
  if (_sock == MAX_SOCK_NUM)
    return;
    ::close(_sock);
    _sock = MAX_SOCK_NUM;
}

void EthernetMaketClient::tick(){
  // Подключаемся к серверу
    switch (connectStatus)
    {
    case CONNECT_START:
        startConnection = millis();
        connectStatus = CONNECT_CONNECTING;
        close();
        Serial.println("\n\nConnecting...");
        break;
    case CONNECT_CONNECTING:
        if (millis() - startConnection >= connectionTimeout)
        {
            connectStatus = CONNECT_FAIL;
            break;
        }
        break;
    case CONNECT_FAIL:
        Serial.print("Failed to connect to server in ");
        Serial.println(isNonBlocking ? "non-blocking mode" : "blocking mode");
        connectStatus = CONNECT_STOP_START;
        break;
    case CONNECT_SUCCESS:
        Serial.print("Connected to server in ");
        Serial.println(isNonBlocking ? "non-blocking mode" : "blocking mode");
        lastActivityTime = millis();
        connectStatus = CONNECT_CONNECTED;

        //! Передача пакетов
        for (size_t i = 0; i < 3; i++)
        {
            client.write((uint8_t *)&packetCounter, sizeof(packetCounter));

            // Печать отправленного пакета
            Serial.print("Sending packet: ");
            for (size_t j = 0; j < sizeof(packetCounter); j++)
            {
                Serial.print(((uint8_t *)&packetCounter)[j], HEX);
                Serial.print(" ");
            }
            Serial.println();
            packetCounter++;
            lastActivityTime = millis(); // Обновляем время последней активности для каждого пакета
        }
        //! Передача завершена

        break;
    case CONNECT_CONNECTED:
        if (!connected())
        {
            Serial.println("Client disconnected");
            connectStatus = CONNECT_STOP_START;
        }

        // Закрытие соединения по таймеру неактивности
        if (connected() && (millis() - lastActivityTime > timeout))
        {
            Serial.println("Connection closed due to inactivity");
            connectStatus = CONNECT_STOP_START;
        }
        break;

    case CONNECT_STOP_START:
        disconnect();
        stopStartTime = millis();
        connectStatus = CONNECT_STOP_WAITING;
    break;

    case CONNECT_STOP_WAITING:
            if (millis() - stopStartTime >= stopTimeout)
            {
                close(); // Закрываем сокет после таймаута
                connectStatus = CONNECT_IDLE;
            }
        break;

    default:
        break;
    }

    switch (connectStatus)
    {
    case CONNECT_START:
    case CONNECT_CONNECTING:
        if (isNonBlocking)
        {
            connectStatus = connectNonBlock(serverIP, SERVER_PORT);
        }
        else
        {
            connectStatus = (ConnectionStatusSimple)connect(serverIP, SERVER_PORT);
        }
        break;

    default:
        break;
    }

    //? Проверка наличия доступных данных от сервера
    if (available() > 0)
    {
        byte response[16];
        int len = read(response, sizeof(response));

        if (len == 16)
        {
            // Проверка контрольной суммы
            uint8_t checksum = 0;
            for (int i = 0; i < 15; i++)
            {
                checksum += response[i];
            }

            Serial.print("RX: ");
            for (int i = 0; i < 16; i++)
            {
                Serial.print(response[i], HEX);
                Serial.print(" ");
            }

            if (response[15] == checksum)
            {
                Serial.println("Checksum is OK");
            }
            else
            {
                Serial.println("Checksum is ERROR");
            }
            lastActivityTime = millis(); // Обновляем время последней активности
        }
        else
        {
            Serial.println("Received an incomplete response");
        }
    }
    //? -------------------------------------------------
}
